// Copyright (c) 2017, the Dart project authors.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

/// Utilities for building CLI apps.
///
/// ## Deprecation notice
///
/// The functionality of this library is incomplete and may be removed in a
/// later version.
///
/// {@category VM}
/// {@nodoc}
@Deprecated(
  "The functionality of this library is incomplete and may be removed in a later version",
)
library dart.cli;

import 'dart:async';
import 'dart:math';
