// Copyright (c) 2013, the Dart project authors.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

/// Helps dealing with reflection in the case that the source code has been
/// changed as a result of compiling with dart2dart.
library dart._mirror_helper;
