<dt id="{{ htmlId }}">
  <span class="name {{ #isDeprecated }}deprecated{{ /isDeprecated }}">{{{ linkedName }}}</span>
  on {{{ extendedElement.linkedName }}}
  {{ >categorization }}
</dt>
<dd>
  {{{ oneLineDoc }}}
</dd>

