// Copyright (c) 2022, the Dart project authors.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

// Prototype js util library for wasm.
library dart.js_util_wasm;

external Object allowInterop<F extends Function>(F f);
