{{#hasSourceHref}}
  <div id="external-links" class="btn-group"><a title="View source code" class="source-link" href="{{{sourceHref}}}"><span class="material-symbols-outlined">description</span></a></div>
{{/hasSourceHref}}
