// Copyright (c) 2012, the Dart project authors.  Please see the AUTHORS file
// for details. All rights reserved. Use of this source code is governed by a
// BSD-style license that can be found in the LICENSE file.

// TODO(http://dartbug.com/45908): Remove empty file after updating Flutter's
// copies of libraries.json and libraries.yaml and waiting for roll into G3.
