{{ >annotations }}

<span class="returntype">{{{ modelType.returnType.linkedName }}}</span>
{{ >name_summary }}{{{ genericParameters }}}(<wbr>{{ #hasParameters }}{{{ linkedParamsLines }}}{{ /hasParameters }})
