<ol>
  {{#localPackages}}
    {{#isFirstPackage}}
      {{#hasDocumentedCategories}}
      <li class="section-title">Topics</li>
        {{#documentedCategoriesSorted}}
        <li>{{{linkedName}}}</li>
        {{/documentedCategoriesSorted}}
      {{/hasDocumentedCategories}}
      <li class="section-title">Libraries</li>
    {{/isFirstPackage}}
    {{^isFirstPackage}}
      <li class="section-title">{{name}}</li>
    {{/isFirstPackage}}
    {{#defaultCategory.publicLibrariesSorted}}
      <li>{{{linkedName}}}</li>
    {{/defaultCategory.publicLibrariesSorted}}
    {{#categoriesWithPublicLibraries}}
      <li class="section-subtitle">{{name}}</li>
      {{#externalItems}}
        <li class="section-subitem">
          <a href="{{url}}" target="_blank">
            {{name}}
            <span class="material-symbols-outlined">open_in_new</span>
          </a>
        </li>
      {{/externalItems}}
      {{#publicLibrariesSorted}}
        <li class="section-subitem">{{{linkedName}}}</li>
      {{/publicLibrariesSorted}}
    {{/categoriesWithPublicLibraries}}
  {{/localPackages}}
</ol>
