{{#hasSourceCode}}
<section class="summary source-code" id="source">
  <h2><span>Implementation</span></h2>
  <pre class="language-dart"><code class="language-dart">{{{ sourceCode }}}</code></pre>
</section>{{/hasSourceCode}}
